import dash
import socket
import multiprocessing
import time
import uuid
import sys
import re
import io
import os
import inspect

from ansi2html import Ansi2HTMLConverter
import requests
from retrying import retry

from werkzeug.debug.tbtools import get_current_traceback

# Get ipython instance
try:
    from IPython import get_ipython
    ipython = get_ipython()
except TimeoutError:
    ipython = None

from .version import __version__

# Find open port
def get_open_port():
    s = socket.socket()
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(('', 0))
    _, port = s.getsockname()
    s.close()
    return port


def pathname_prefix(proxy_port, workspace_id, cluster_id):
    return '/driver-proxy/o/{workspace_id}/{cluster_id}/{proxy_port}' \
        .format(workspace_id=workspace_id,
                cluster_id=cluster_id,
                proxy_port=proxy_port)


def proxy_uri(host, proxy_port, workspace_id, cluster_id):  # format proxy URI
    return 'https://{workspace_host}{pathname_prefix}' \
        .format(workspace_host=host,
                pathname_prefix=pathname_prefix(proxy_port, workspace_id, cluster_id))


class DatabricksDash(dash.Dash):
    """A Dash subclass for developing Dash apps interactively in Databricks notebooks

    :param host: Hostname of the Databricks instance
    :param workspace_id: Databricks workspace id
    :param workspace_id: Databricks cluster id

    See parent docstring for additional parameters
    """
    server_processes = {}
    _token = str(uuid.uuid4())
    _dbutils = None
    _default_port = None

    def __init__(self, name, **kwargs):
        super(DatabricksDash, self).__init__(name, **kwargs)

        # Initialize properties
        self.port = None
        self.cluster_id = None
        self.workspace_id = None
        self.host = None

        if ipython is None:
            # Not running in a notebook
            return

        # Search call stack for the Databricks dbutils object
        self._find_dbutils()

        # Get cluser ID, workspace ID, and hostname using dbutils
        db_context = self._dbutils.notebook.entry_point.getDbutils().notebook().getContext()
        db_context_tags = db_context.tags()
        self.cluster_id = db_context.clusterId().get()
        self.workspace_id = db_context_tags.get("orgId").get()
        self.host = db_context_tags.get("browserHostName").get()

        # Register route that we can use to poll to see when server is running
        @self.server.route('/_alive_' + DatabricksDash._token, methods=['GET'])
        def alive():
            return 'Alive'

    @classmethod
    def _find_dbutils(cls):
        import dbutils as dbutils_module
        if cls._dbutils is None:
            for frame in inspect.getouterframes(inspect.currentframe()):
                global_names = set(frame.frame.f_globals)
                # Check for dbutils plus a few others to reduce chance of a false
                # hit.
                if "dbutils" in global_names:
                    dbutils = frame.frame.f_globals["dbutils"]
                    if isinstance(dbutils, dbutils_module.DBUtils):
                        # Found it
                        cls._dbutils = dbutils
                        break

            if cls._dbutils is None:
                raise EnvironmentError(
                    """
Unable to detect the Databricks dbutils object. The 'DatabricksDash' class is only
supported from within a Databricks notebook environment."""
                )

    @classmethod
    def _get_default_port(cls):
        if cls._default_port is None:
            if ipython is None:
                # Not in a notebook context
                cls._default_port = "8050"
            else:
                # In Databricks notebook, find an available port
                cls._default_port = get_open_port()
        return cls._default_port

    def _build_prefix(self, port):
        return pathname_prefix(port, self.workspace_id, self.cluster_id)

    def run_server(
            self, mode='external', width=1000, height=800, **kwargs
    ):
        """
        Serve the app using flask in a forked process. You should not run this on a
        production server, use gunicorn/waitress instead.

        :param mode: Display mode. One of:
            ``"external"``: The URL of the app will be displayed in the notebook
                output cell. Clicking this URL will open the app in the default
                web browser.
            ``"inline"``: The app will be displayed inline in the notebook output cell
                in an iframe.
        :param width: Width of app when displayed using mode="inline"
        :param height: Height of app when displayed using mode="inline"
        :param kwargs: Additional keyword arguments to pass to the superclass
            ``Dash.run_server`` method.
        """
        # Get superclass run_server method
        super_run_server = super(DatabricksDash, self).run_server

        if ipython is None:
            # Not in IPython context, call run run_server synchronously
            super_run_server(**kwargs)
            return

        # Validate / infer display mode
        valid_display_modes = ["inline", "external"]
        if not isinstance(mode, str):
            raise ValueError(
                "The mode argument must be a string\n"
                "    Received value of type {typ}: {val}".format(
                    typ=type(mode), val=repr(mode)
                )
            )
        else:
            mode = mode.lower()
            if mode not in valid_display_modes:
                raise ValueError(
                    "Invalid mode argument {mode}\n"
                    "    Valid arguments: {valid_display_modes}".format(
                        mode=repr(mode), valid_display_modes=valid_display_modes
                    )
                )

        # Get host and port
        host = kwargs.get("host", os.getenv("HOST", "0.0.0.0"))
        port = kwargs.get("port", os.getenv("PORT", self._get_default_port()))

        kwargs['host'] = host
        kwargs['port'] = port

        # Save current port
        self.port = port

        # Terminate any existing server process
        self.terminate_server()

        # Update app config
        self.config.update({
            'requests_pathname_prefix': self._build_prefix(port),
            'eager_loading': True,
        })

        # Process dev tools
        # Default the global "debug" flag to True
        debug = kwargs.pop('debug', True)

        # Enable supported dev tools
        if debug:
            for k in [
                'dev_tools_silence_routes_logging',
                'dev_tools_props_check',
                'dev_tools_serve_dev_bundles',
                'dev_tools_prune_errors',
                'dev_tools_ui',
            ]:
                if k not in kwargs:
                    kwargs[k] = True

            if 'dev_tools_hot_reload' not in kwargs:
                # Enable hot-reload by default in "external" mode. Enabling when inline
                # can cause Jupyter problems sometimes when there is no active kernel.
                kwargs['dev_tools_hot_reload'] = mode == "external"

        # Set up custom callback exception handling
        self._config_callback_exception_handling(
            dev_tools_prune_errors=kwargs.get('dev_tools_prune_errors', True),
        )

        def run():
            super_run_server(**kwargs)

        server_process = multiprocessing.Process(
            target=run,
        )

        server_process.daemon = True

        dashboard_url = proxy_uri(self.host, port, self.workspace_id, self.cluster_id)

        server_process.start()
        self.server_processes[port] = server_process

        # Wait for server to start up
        alive_url = "http://{host}:{port}/_alive_{token}".format(
            host='0.0.0.0', port=port, token=DatabricksDash._token
        )

        # Wait for app to respond to _alive endpoint
        @retry(
            stop_max_attempt_number=15,
            wait_exponential_multiplier=10,
            wait_exponential_max=1000
        )
        def wait_for_app():
            res = requests.get(alive_url).content.decode()
            if res != "Alive":
                url = "http://{host}:{port}".format(
                    host=self.host, port=port,
                )
                raise OSError(
                    "Address '{url}' already in use.\n"
                    "    Try passing a different port to run_server.".format(
                        url=url
                    )
                )

        wait_for_app()

        from IPython.display import HTML, IFrame
        display = self._dbutils.shell.display
        if mode == "inline":
            display(IFrame(dashboard_url, width, height))
        else:
            display(HTML(
                "Dash app running on "
                "<a href='{dashboard_url}'>{dashboard_url}</a>".format(
                    dashboard_url=dashboard_url
                )
            ))

    def _config_callback_exception_handling(
            self, dev_tools_prune_errors,
    ):

        @self.server.errorhandler(Exception)
        def _wrap_errors(_):
            from IPython.core.ultratb import FormattedTB

            """Install traceback handling for callbacks"""
            self._traceback = sys.exc_info()[2]

            # Compute number of stack frames to skip to get down to callback
            tb_werkzeug = get_current_traceback()
            skip = 0
            if dev_tools_prune_errors:
                for i, line in enumerate(tb_werkzeug.plaintext.splitlines()):
                    if "%% callback invoked %%" in line:
                        skip = int((i + 1) / 2)
                        break

            # Use IPython traceback formatting to build colored ANSI traceback
            # string
            ostream = io.StringIO()
            ipytb = FormattedTB(
                tb_offset=skip,
                mode="Verbose",
                color_scheme="Linux",
                include_vars=True,
                ostream=ostream
            )
            ipytb()

            # Print colored ANSI representation if requested
            ansi_stacktrace = ostream.getvalue()

            # Use ansi2html to convert the colored ANSI string to HTML
            conv = Ansi2HTMLConverter(scheme="ansi2html", dark_bg=False)
            html_str = conv.convert(ansi_stacktrace)

            # Set width to fit 75-character wide stack trace and font to a size the
            # won't require a horizontal scroll bar
            html_str = html_str.replace(
                '<html>',
                '<html style="width: 75ch; font-size: 0.86em">'
            )

            # Remove explicit background color so Dash dev-tools can set background
            # color
            html_str = re.sub("background-color:[^;]+;", "", html_str)

            return html_str, 500

    def terminate_server(self):
        if isinstance(self.server_processes.get(self.port, None),
                      multiprocessing.Process):
            try:
                self.server_processes[self.port].terminate()
            except Exception:
                pass

            # Sleep briefly to let server finish shutting down
            time.sleep(0.5)
